export interface IMap<K, V> {
    root: ITrieNode<K, V>;
    size: number;
}
export declare type Key = number | string;
export interface INumberIndexable<K, V> {
    [key: number]: ISingleValueNode<K, V>;
}
export interface IStringIndexable<K, V> {
    [key: string]: ISingleValueNode<K, V>;
}
export interface ITrieNode<K, V> {
    [index: number]: ITrieNode<K, V> | IMultiValueNode<K, V> | ISingleValueNode<K, V>;
    length: number;
}
export interface IMultiValueNode<K, V> {
    map: INumberIndexable<K, V> & IStringIndexable<K, V>;
}
export interface ISingleValueNode<K, V> {
    key: K;
    value: V;
}
/**
 * This adapter class can be used to interact with an Immerutable Map stored in the ngrx store.
 * Immerutable maps are very similar to maps provided by ImmutableJS. They use a trie structure
 * for structural sharing which allows for items to be inserted and removed without the need
 * to shallow copy the entire map. The keys used by this map may be strings or numbers only.
 *
 * Runtimes:
 * Get/Has: O(1)
 * Set: O(1)
 * Remove: O(1)
 * Iterate: O(n)
 * Note: The constant factor for these operations will be considerably higher than for standard maps.
 */
export declare class MapAdapter<K extends Key, V> {
    /** The number of bits to use per level of the trie. */
    protected shift: number;
    /** The maximum length of an internal node in the map (containing value and child pointers). */
    protected trieNodeSize: number;
    /** The mask used to grab the last "shift" bits from the key. */
    protected mask: number;
    /** The maximum number of levels in the tree (when we've used up all the bits in the key). */
    protected maxDepth: number;
    /**
     * Creates a new Immerutable map. This map should be stored in the ngrx store.
     * It should not be modified or read from directly. All interaction with this
     * map should happen through this adapter class.
     */
    create(): IMap<K, V>;
    /** Returns true if the map contains the key, false otherwise. */
    has(map: IMap<K, V>, key: K): boolean;
    /** Gets the value for the specified key. If the key does not exist, undefined is returned. */
    get(map: IMap<K, V>, key: K): V | undefined;
    /**
     * Stores the specified value in the map using the specified key.
     * If the key already exists, it will be replaced with the new value.
     */
    set(map: IMap<K, V>, key: K, value: V): void;
    /**
     * Removes the specified key from the map. If the key does not exist, this is a no-op.
     */
    remove(map: IMap<K, V>, key: K): void;
    /**
     * Updates the value of the specified key in the map using an updater function.
     * The updater function will receive the existing value, and may either mutate it directly
     * or return a new value, which will take its place.
     * If the key is not found, the updater function is not called and no update is made.
     *
     * Example:
     * ```
     * const adapter = new MapAdapter<number, { data: string }>();
     * const map = adapter.create();
     *
     * adapter.set(map, 1, { data: 'one' });
     * adapter.update(map, 1, (value) => { value.data = 'ichi'; });
     * ```
     */
    update(map: IMap<K, V>, key: K, updater: (item: V) => V | void | undefined): V | undefined;
    /** Gets the number of keys in the map. */
    getSize(map: IMap<K, V>): number;
    /**
     * Returns an Iterable which can be used to iterate through all the keys & values in the map.
     * Order of iteration is not guaranteed, however it will be consistent across multiple iterations.
     * DO NOT add or remove items from the map while iterating!
     *
     * This method requires Symbol.iterator or a polyfill. If using Typescript and compiling to ES5, the
     * downlevelIteration setting is recommended. Note that the iterable can be passed to Array.from to convert
     * to an array (note that this incurs a larger performance and memory cost compared to just iterating).
     *
     * Example (ES6 or ES5 w/ downlevelIteration):
     * ```
     * const adapter = new MapAdapter<string, T>();
     * ...
     *
     * for ({key, value} of adapter.getIterable(map)) {
     *   console.log(key, value);
     * }
     * ```
     *
     * Example (old school):
     * ```
     * const adapter = new MapAdapter<string, T>();
     * ...
     *
     * const iterable = adapter.getIterable(map);
     * const iterator = iterable[Symbol.iterator](); // May need Symbol.iterator polyfill
     * let next: T;
     * while (!(next = iterator.next()).done) {
     *   const {key, value} = next.value;
     *   console.log(key, value);
     * }
     * ```
     */
    getIterable(map: IMap<K, V>): Iterable<ISingleValueNode<K, V>>;
    private createTrieNode;
    private createSingleValueNode;
    private createValueNode;
    private lookupValueNode;
    private pushSingleValueNodeDown;
    private computePartialHashCode;
}
