export interface IBTreeNode<T> {
    isRoot?: boolean;
    items: IBTreeValueNode<T>[];
    children?: Array<IBTreeNode<T>>;
}
export interface ISortedCollection<T> {
    root: IBTreeNode<T>;
    size: number;
}
export interface IBTreeValueNode<T> {
    value: T;
}
export declare type ParentPath<T> = {
    index: number;
    node: IBTreeNode<T>;
}[];
export interface LookupNodeInfo<T> {
    valueNode: IBTreeValueNode<T>;
    parentPath: ParentPath<T>;
}
export declare type Comparer<T> = (a: T, b: T) => number;
export declare type EqualityComparer<T> = (a: T, b: T) => boolean;
/**
 * This adapter class can be used to interact with an Immerutable Sorted Collection stored in the ngrx store.
 * This collection is backed by a B-tree which can efficiently handle insertions and deletions while maintaining
 * the sorted order of the collection. B-trees also allow for structural sharing such that only a portion
 * of the data structure needs to be shallow copied when items are added or removed.
 *
 * This data structure allows duplicates and does not provide key-based lookups. A SortedMap should be used instead
 * if these properties are desirable.
 *
 * Runtimes:
 * Insert: O(log(n))
 * Remove: O(log(n))
 * Iterate: O(n)
 */
export declare class SortedCollectionAdapter<T> {
    orderComparer: Comparer<T>;
    equalityComparer: EqualityComparer<T>;
    private maxItemsPerLevel;
    private minItemsPerLevel;
    /**
     * @param args.orderComparer A function which compares two values (similar to Array.sort). Used for ordering this collection.
     * @param args.equalityComparer A function which tests two values for equality. Optional (by default uses ===).
     * @param args.maxItemsPerLevel Don't set this value; it is only available for use in unit testing.
     */
    constructor(args: {
        orderComparer: Comparer<T>;
        equalityComparer?: EqualityComparer<T>;
        maxItemsPerLevel?: number;
    });
    /**
     * Creates a new Immerutable Sorted Collection. This collection should be stored in the ngrx store.
     * It should not be modified or read from directly. All interaction with this collection should
     * happen through this adapter class.
     */
    create(): ISortedCollection<T>;
    /** Inserts an item into the collection in sorted order. */
    insert(collection: ISortedCollection<T>, value: T): void;
    getSize(collection: ISortedCollection<T>): number;
    update(collection: ISortedCollection<T>, value: T, updater: (item: T) => T | void | undefined): void | T;
    /**
     * Ensures that a value is still in sorted order after being mutated.
     * In general, prefer using the "update" method instead.
     *
     * @param nodeInfo The return value of of calling lookupValuePath for the value being mutated.
     */
    ensureSortedOrderOfNode(collection: ISortedCollection<T>, nodeInfo: LookupNodeInfo<T>): void;
    lookupValuePath(collection: ISortedCollection<T>, value: T): LookupNodeInfo<T> | undefined;
    getFirst(collection: ISortedCollection<T>): T | undefined;
    getLast(collection: ISortedCollection<T>): T | undefined;
    remove(collection: ISortedCollection<T>, value: T): void;
    getIterable(collection: ISortedCollection<T>): Iterable<T>;
    private insertInBTreeNode;
    private isLeafNode;
    private splitNode;
    private splitLeafNodeLeftHeavy;
    private splitLeafNodeRightHeavy;
    private findLeafNodeInsertionPoint;
    private findRecursionIndex;
    private removeByPath;
    private rebalance;
    private isNodeDeficient;
    private canNodeLoseItem;
    private lookupLeftMostValueWithParentPath;
    private lookupRightMostValueWithParentPath;
    private getFurthestLeftValue;
    private getFurthestRightValue;
    private getPreviousValue;
    private getNextValue;
    private _lookupValuePath;
    private binarySearchForInsertion;
    private binarySearchForLookup;
    private _binarySearch;
    private createBTreeNode;
    private createBTreeRootNode;
    private createBTreeValueNode;
}
